# """Console script for fintorch."""

import click
import optuna
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.callbacks.early_stopping import EarlyStopping
from lightning.pytorch.cli import LightningCLI
from optuna.integration import WeightsAndBiasesCallback

from fintorch.datasets import elliptic as e
from fintorch.datasets import ellipticpp as epp
from fintorch.models.graph.graphbean.wand.sweep import objective


@click.group()
def fintorch():
    """FinTorch CLI - Your financial AI toolkit"""
    pass


@fintorch.command()
@click.argument("dataset")
def datasets(dataset):
    """Download financial datasets"""
    # Implement your dataset download logic here
    click.echo(f"Downloading dataset: {dataset}")
    if dataset == "elliptic":
        e.TransactionDataset("~/.fintorch_data", force_reload=True)
    elif dataset == "ellipticpp":
        epp.TransactionActorDataset("~/.fintorch_data", force_reload=True)


@fintorch.command()
@click.option('--model', required=True, help='Name of the model to use.')
@click.option('--predict',
              required=True,
              help='Type of prediction to perform.')
@click.option('--max_epochs', required=False, help='Max epochs')
def sweep(model, predict, max_epochs):
    """Sweep your financial models"""
    # Implement your model training logic here
    click.echo(f"Starting sweep for model: {model}")

    if model == "graphbean_elliptic":
        wandb_kwargs = {"project": f"graphbean-{predict}"}
        wandbc = WeightsAndBiasesCallback(wandb_kwargs=wandb_kwargs,
                                          as_multirun=True)

        study = optuna.create_study(direction="maximize",
                                    pruner=optuna.pruners.MedianPruner())

        def wrapped_objective(trial):
            return objective(trial, max_epochs, predict)

        # study.optimize(objective, n_trials=100)
        study.optimize(wrapped_objective, n_trials=100, callbacks=[wandbc])

        print("Number of finished trials: {}".format(len(study.trials)))

        print("Best trial:")
        trial = study.best_trial

        print("  Value: {}".format(trial.value))

        print("  Params: ")
        for key, value in trial.params.items():
            print("    {}: {}".format(key, value))


def trainer():
    cli = LightningCLI(
        run=False,
        save_config_callback=None,
        trainer_defaults={
            "max_epochs":
            10,
            "callbacks": [
                ModelCheckpoint(monitor="val_loss", mode="min"),
                EarlyStopping(monitor="train_loss", mode="min"),
            ],
        },
    )

    cli.trainer.fit(cli.model, datamodule=cli.datamodule)



from typing import List, Optional, Union

import torch
from torch import Tensor
from torch_geometric.nn.aggr import Aggregation
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.nn.dense.linear import Linear
from torch_geometric.typing import Adj, OptPairTensor, Size


class BeanAggregation(Aggregation):
    """
    BeanAggregation class performs node and edge aggregation operations.

    Reference:

    Fathony, Rizal, Jenn Ng, and Jia Chen. 2023. “Interaction-Focused Anomaly Detection on Bipartite
    Node-and-Edge-Attributed Graphs.” In 2023 International Joint Conference on Neural Networks (IJCNN), 1–10. IEEE.

    Args:
        x (Tensor): The input tensor.
        index (Optional[Tensor]): The index tensor. Default is None.
        ptr (Optional[Tensor]): The pointer tensor. Default is None.
        dim_size (Optional[int]): The size of the dimension. Default is None.
        dim (int): The dimension along which to perform the aggregation. Default is -2.
        edge_attr: The edge attribute tensor. Default is None.
        **kwargs: Additional keyword arguments.

    Returns:
        Tensor: The aggregated output tensor.

    """

    def forward(
        self,
        x: Tensor,
        index: Optional[Tensor] = None,
        ptr: Optional[Tensor] = None,
        dim_size: Optional[int] = None,
        dim: int = -2,
        edge_attr=None,
        **kwargs,
    ) -> Tensor:
        """
        Forward pass of the BeanAggregation layer.

        Args:
            x (Tensor): The input tensor.
            index (Optional[Tensor]): The index tensor. Default is None.
            ptr (Optional[Tensor]): The pointer tensor. Default is None.
            dim_size (Optional[int]): The size of the dimension. Default is None.
            dim (int): The dimension along which to perform the aggregation. Default is -2.
            edge_attr: The edge attribute tensor. Default is None.
            **kwargs: Additional keyword arguments.

        Returns:
            Tensor: The aggregated output tensor.

        """

        # Node aggregation
        output_mean = self.reduce(x, index, ptr, dim_size, dim, reduce="mean")
        output_sum = self.reduce(x, index, ptr, dim_size, dim, reduce="max")

        # Edge aggregation
        output_edge_mean = self.reduce(edge_attr,
                                       index,
                                       ptr,
                                       dim_size,
                                       dim,
                                       reduce="mean")
        output_edge_sum = self.reduce(edge_attr,
                                      index,
                                      ptr,
                                      dim_size,
                                      dim,
                                      reduce="max")

        # Concatenate outputs
        output = torch.cat(
            (output_mean, output_sum, output_edge_mean, output_edge_sum),
            dim=1)

        return output


class BeanAggregationSimple(Aggregation):
    """
    A simple bean aggregation layer (NOTE: only use this one if you don't have edge attributes).

    This layer performs node aggregation by calculating the mean and maximum values
    along a specified dimension of the input tensor. The outputs are then concatenated
    along the same dimension. Use this Aggregator if you don't have edge features!

    Reference:

    Fathony, Rizal, Jenn Ng, and Jia Chen. 2023. “Interaction-Focused Anomaly Detection on Bipartite
    Node-and-Edge-Attributed Graphs.” In 2023 International Joint Conference on Neural Networks (IJCNN), 1–10. IEEE.

    Args:
        x (Tensor): The input tensor.
        index (Optional[Tensor]): The index tensor for sparse input. Default is None.
        ptr (Optional[Tensor]): The pointer tensor for sparse input. Default is None.
        dim_size (Optional[int]): The size of the dimension along which to perform aggregation.
            Default is None.
        dim (int): The dimension along which to perform aggregation. Default is -2.
        **kwargs: Additional keyword arguments.

    Returns:
        Tensor: The concatenated output tensor.

    """

    def forward(
        self,
        x: Tensor,
        index: Optional[Tensor] = None,
        ptr: Optional[Tensor] = None,
        dim_size: Optional[int] = None,
        dim: int = -2,
        **kwargs,
    ) -> Tensor:
        """
        Forward pass of the bean aggregation layer.

        Args:
            x (Tensor): The input tensor.
            index (Optional[Tensor]): The index tensor for sparse input. Default is None.
            ptr (Optional[Tensor]): The pointer tensor for sparse input. Default is None.
            dim_size (Optional[int]): The size of the dimension along which to perform aggregation.
                Default is None.
            dim (int): The dimension along which to perform aggregation. Default is -2.
            **kwargs: Additional keyword arguments.

        Returns:
            Tensor: The concatenated output tensor.

        """

        # Node aggregation
        output_mean = self.reduce(x, index, ptr, dim_size, dim, reduce="mean")
        output_sum = self.reduce(x, index, ptr, dim_size, dim, reduce="max")

        # Concatenate outputs
        output = torch.cat((output_mean, output_sum), dim=1)

        return output


class BEANConvSimple(MessagePassing):
    """
    BEANConvSimple is a graph convolutional layer. In comparison with the BEANConv layer, this layer does not
    use edge attributes.

    Reference:

    Fathony, Rizal, Jenn Ng, and Jia Chen. 2023. “Interaction-Focused Anomaly Detection on Bipartite
    Node-and-Edge-Attributed Graphs.” In 2023 International Joint Conference on Neural Networks (IJCNN), 1–10. IEEE.

    Args:
        in_channels (int): The number of input channels.
        out_channels (int): The number of output channels.
        bias (bool, optional): If set to True, enables the learnable bias. Default is True.
        normalize (bool, optional): If set to True, applies batch normalization to the output. Default is True.
        node_self_loop (bool, optional): If set to True, includes the node's own features in the output.
        Default is True.
        aggr (str, List[str], Aggregation, optional): The aggregation method to use. Default is "mean".
        **kwargs: Additional keyword arguments to be passed to the parent class.

    Attributes:
        in_channels (int): The number of input channels.
        out_channels (int): The number of output channels.
        bias (bool): If True, enables the learnable bias.
        normalize (bool): If True, applies batch normalization to the output.
        node_self_loop (bool): If True, includes the node's own features in the output.
        lin_l (Linear): Linear layer for transforming the output.
        node_normalize (BatchNorm1d): Batch normalization layer for normalizing the output.

    Methods:
        reset_parameters(): Resets the parameters of the layer.
        forward(x, edge_index, size=None): Performs the forward pass of the layer.
        __repr__(): Returns a string representation of the layer.

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        bias: int = True,
        normalize: bool = True,
        node_self_loop: bool = True,
        aggr: Optional[Union[str, List[str], Aggregation]] = "mean",
        **kwargs,
    ):
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.bias = bias
        self.normalize = normalize
        self.node_self_loop = node_self_loop

        aggr = BeanAggregationSimple()

        super().__init__(aggr, **kwargs)
        self.lin_l = Linear(-1, out_channels, bias=bias)

        if self.normalize:
            self.node_normalize = torch.nn.BatchNorm1d(out_channels)

        self.reset_parameters()

    def reset_parameters(self):
        """
        Resets the parameters of the layer.
        """
        super().reset_parameters()
        self.lin_l.reset_parameters()
        if self.normalize:
            self.node_normalize.reset_parameters()

    def forward(
        self,
        x: Union[Tensor, OptPairTensor],
        edge_index: Adj,
        size: Size = None,
    ):
        """
        Performs the forward pass of the layer.

        Args:
            x (Tensor or OptPairTensor): The input node features.
            edge_index (Adj): The adjacency matrix of the graph.
            size (Size, optional): The size of the graph. Default is None.

        Returns:
            Tensor: The output node features.

        """
        output = self.propagate(edge_index, x=x)
        if self.node_self_loop:
            output = torch.cat((x[1], output), dim=1)

        output = self.lin_l(output)

        if self.normalize:
            output = self.node_normalize(output)

        return output

    def __repr__(self) -> str:
        """
        Returns a string representation of the layer.

        Returns:
            str: The string representation of the layer.

        """
        return (f"{self.__class__.__name__}({self.in_channels}, "
                f"{self.out_channels}, aggr={self.aggr})")


class BEANConv(MessagePassing):
    r"""
    BEANConv is a message-passing graph convolutional layer implementation.

    Reference:

    Fathony, Rizal, Jenn Ng, and Jia Chen. 2023. “Interaction-Focused Anomaly Detection on Bipartite
    Node-and-Edge-Attributed Graphs.” In 2023 International Joint Conference on Neural Networks (IJCNN), 1–10. IEEE.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        edge_in_channels (int, optional): Number of input channels for edge features. Default is -1.
        edge_out_channels (int, optional): Number of output channels for edge features. Default is -1.
        bias (bool, optional): If set to True, enables the learnable bias. Default is True.
        normalize (bool, optional): If set to True, applies batch normalization to node and edge features.
        Default is True.
        node_self_loop (bool, optional): If set to True, includes self-loops in the graph. Default is True.
        aggr (str, List[str], Aggregation, optional): The aggregation method to use for message passing.
        Default is "mean".
        has_edge_features (bool, optional): If set to True, indicates that the graph has edge features. Default is True.

    Attributes:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        bias (bool): If True, enables the learnable bias.
        normalize (bool): If True, applies batch normalization to node and edge features.
        node_self_loop (bool): If True, includes self-loops in the graph.
        edge_out_channels (int): Number of output channels for edge features.
        edge_in_channels (int): Number of input channels for edge features.
        has_edge_features (bool): If True, indicates that the graph has edge features.

    Methods:
        reset_parameters(): Resets the parameters of the BEANConv layer.
        forward(x, edge_index, size, edge_attr): Performs a forward pass of the BEANConv layer.
        aggregate(inputs, index, ptr, edge_attr, dim_size): Aggregates the input features according to
        the specified aggregation method.
        edge_update(x_j, x_i): Updates the edge features based on the input node features.
        __repr__(): Returns a string representation of the BEANConv layer.

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        edge_in_channels: int = -1,
        edge_out_channels: int = -1,
        bias: int = True,
        normalize: bool = True,
        node_self_loop: bool = True,
        aggr: Optional[Union[str, List[str], Aggregation]] = "mean",
        has_edge_features=True,
        **kwargs,
    ):
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.bias = bias
        self.normalize = normalize
        self.node_self_loop = node_self_loop
        self.edge_out_channels = edge_out_channels
        self.edge_in_channels = edge_in_channels

        self.has_edge_features = has_edge_features

        aggr = BeanAggregation()

        super().__init__(aggr, **kwargs)
        self.lin_l = Linear(-1, out_channels, bias=bias)

        if self.normalize:
            self.node_normalize = torch.nn.BatchNorm1d(out_channels)
            self.edge_normalize = torch.nn.BatchNorm1d(edge_in_channels)

        if edge_out_channels != -1:
            self.edge_layer = Linear(-1, edge_out_channels, bias=bias)
            self.edge_normalize = torch.nn.BatchNorm1d(edge_out_channels)

        self.reset_parameters()

    def reset_parameters(self):
        r"""
        Resets the parameters of the BEANConv layer.
        """
        super().reset_parameters()
        self.lin_l.reset_parameters()
        if self.normalize:
            self.node_normalize.reset_parameters()
            self.edge_normalize.reset_parameters()

        if self.edge_out_channels != -1:
            self.edge_layer.reset_parameters()

    def forward(
        self,
        x: Union[Tensor, OptPairTensor],
        edge_index: Adj,
        size: Size = None,
        edge_attr=None,
    ):
        r"""
        Performs a forward pass of the BEANConv layer.

        Args:
            x (Tensor or OptPairTensor): Node feature matrix or a tuple of node feature matrices.
            edge_index (Adj): Graph connectivity in COO format.
            size (Size, optional): The size of the graph. Default is None.
            edge_attr (Tensor, optional): Edge feature matrix. Default is None.

        Returns:
            output (Tensor): The output node feature matrix.
            edge_attr (Tensor): The output edge feature matrix.

        """
        # propagate node information
        output = self.propagate(edge_index, x=x, edge_attr=edge_attr)
        # propagate edge information
        edge_attr = self.edge_updater(edge_index, x=x, edge_attr=edge_attr)

        if self.node_self_loop:
            # Add the node's own features
            output = torch.cat((x[1], output), dim=1)

        # Node projection
        output = self.lin_l(output)

        # Edge projection
        if self.edge_out_channels != -1 and edge_attr is not None:
            edge_attr = self.edge_layer(edge_attr)

        if self.normalize:
            output = self.node_normalize(output)
            edge_attr = self.edge_normalize(edge_attr)

        return output, edge_attr

    def aggregate(
        self,
        inputs: Tensor,
        index: Tensor,
        ptr: Optional[Tensor] = None,
        edge_attr: Tensor = None,
        dim_size: Optional[int] = None,
    ) -> Tensor:
        r"""
        Aggregates the input features according to the specified aggregation method, in this case we added
        the edge_attr as an additional parameter to facilitate edge message passing to nodes.

        Args:
            inputs (Tensor): The input features to aggregate.
            index (Tensor): The indices of the nodes to aggregate.
            ptr (Tensor, optional): The pointer tensor pointing to the start and end indices of the nodes.
                Default is None.
            edge_attr (Tensor, optional): The edge features to aggregate. Default is None.
            dim_size (int, optional): The size of the dimension to aggregate. Default is None.

        Returns:
            Tensor: The aggregated output features.

        """
        # Overwrite standard aggr_module call
        return self.aggr_module(inputs,
                                index,
                                ptr,
                                dim_size,
                                edge_attr=edge_attr)

    def edge_update(self, x_j: Tensor, x_i: Tensor) -> Tensor:
        r"""
        Updates the edge features based on the input node features.

        Args:
            x_j (Tensor): The input features of the neighboring nodes.
            x_i (Tensor): The input features of the central nodes.

        Returns:
            Tensor: The updated edge features.

        """
        return torch.cat((x_i, x_j), dim=1)

    def __repr__(self) -> str:
        r"""
        Returns a string representation of the BEANConv layer.

        Returns:
            str: The string representation of the BEANConv layer.

        """
        return (f"{self.__class__.__name__}({self.in_channels}, "
                f"{self.out_channels}, aggr={self.aggr})")



import os
from typing import Callable, List, Optional

import numpy as np
import polars as pol
import torch
from torch_geometric.data import Data, InMemoryDataset

from fintorch.datasets.kaggle.downloader import KaggleDownloader


class TransactionDataset(InMemoryDataset):
    """
    The Elliptic Data Set: Understanding Bitcoin Transactions

    The Elliptic Data Set provides valuable insights into Bitcoin transactions,
    allowing researchers to distinguish between licit and illicit activities.
    Here's what you need to know:

    Purpose:

    - Maps Bitcoin transactions to real-world entities.
    - Classifies entities as licit or illicit.

    Content:

    - Anonymized transaction graph from the Bitcoin blockchain.
    - Nodes represent transactions; edges represent Bitcoin flows.
    - 203,769 nodes and 234,355 edges.
    - Each node has 166 features.
    - Nodes are labeled "licit", "illicit", or "unknown".

    Features:

    - Include transaction time, inputs/outputs, fees, volume, and aggregated statistics.
    - Time steps group transactions into connected components.

    Citations:

    If you use the Elliptic Data Set, please cite:
    [1] Elliptic, www.elliptic.co.
    [2] M. Weber, et al., "Anti-Money Laundering in Bitcoin:
    Experimenting with Graph Convolutional Networks for Financial Forensics",
    KDD ’19 Workshop on Anomaly Detection in Finance, August 2019, Anchorage, AK, USA.

    The EllipticDataset consists of three files:

    - elliptic_txs_classes.csv
    - elliptic_txs_edgelist.csv
    - elliptic_txs_features.csv

    Parameters:
    - root (str): Root directory where the dataset exists or will be downloaded.
    - transform (Optional[Callable]): A function/transform that takes in an
      object and returns a transformed version. Default is None.
    - pre_transform (Optional[Callable]): A function/transform that takes in an
      object and returns a transformed version. Default is None.
    - pre_filter (Optional[Callable]): A function that takes in an object and
      returns a boolean value indicating whether the object should be included
      in the final dataset. Default is None.
    - force_reload (bool): Flag indicating whether to force reloading the dataset
      even if it already exists. Default is False.

    Returns:
    None
    """

    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        pre_transform: Optional[Callable] = None,
        pre_filter: Optional[Callable] = None,
        force_reload: bool = False,
    ) -> None:
        super().__init__(root,
                         transform,
                         pre_transform,
                         pre_filter,
                         force_reload=force_reload)

        self.load(self.processed_paths[0])
        data = self.get(0)
        self.data, self.slices = self.collate([data])

        assert isinstance(self._data, Data)

    @property
    def raw_file_names(self):
        """
        Returns a list of raw file names for the elliptic dataset.

        Returns:
            list: A list of raw file names.
        """
        return [
            "elliptic_bitcoin_dataset/elliptic_txs_features.csv",
            "elliptic_bitcoin_dataset/elliptic_txs_edgelist.csv",
            "elliptic_bitcoin_dataset/elliptic_txs_classes.csv",
        ]

    @property
    def processed_file_names(self) -> List[str]:
        """
        Returns a list of processed file names.

        Returns:
            list: A list of processed file names.
        """
        return ["transaction_graph_v1.pt"]

    def download(self) -> None:
        """
        Downloads the elliptic dataset from Kaggle to `self.raw_dir`.
        """
        # Download from Kaggle to `self.raw_dir`.
        downloader = KaggleDownloader()
        downloader.download_dataset("ellipticco/elliptic-data-set",
                                    download_dir=self.raw_dir)

    def process(self) -> None:
        """
        Process the dataset by reading the required files, performing data transformations,
        and saving the processed data to disk.

        Returns:
            None
        """

        data_list = []

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        features_file_path = os.path.join(self.raw_dir,
                                          "elliptic_bitcoin_dataset",
                                          "elliptic_txs_features.csv")

        edgelist_file_path = os.path.join(self.raw_dir,
                                          "elliptic_bitcoin_dataset",
                                          "elliptic_txs_edgelist.csv")
        classes_file_path = os.path.join(self.raw_dir,
                                         "elliptic_bitcoin_dataset",
                                         "elliptic_txs_classes.csv")

        features = pol.read_csv(features_file_path, has_header=False)
        edgelist = pol.read_csv(edgelist_file_path)
        classes = pol.read_csv(classes_file_path)

        # Mapping 'class' column to numerics
        # The dataset has licit (0), illicit (1), and unknown (2) entities.
        classes = classes.with_columns(
            pol.col("class").cast(pol.Utf8).map_elements(
                lambda x: {
                    "unknown": 2,
                    "1": 1,
                    "2": 0
                }.get(x),
                return_dtype=pol.Int64))

        # Cast the classes to Int32 identifiers
        classes = classes.with_columns(classes["txId"].cast(
            pol.Int64).alias("txId"))

        # Merging DataFrames efficiently with Polars
        # The resulting DataFrame consists of the classes for each node
        df_merge = features.join(classes,
                                 how="left",
                                 left_on="column_1",
                                 right_on="txId")

        # Drop redundant column
        df_merge = df_merge.drop("txId")

        nodes = df_merge.select("column_1").to_numpy().flatten()
        self.map_id = dict(zip(nodes, range(len(nodes))))

        edgelist = edgelist.with_columns(
            txId1=pol.col("txId1").replace(self.map_id, default=-1))

        edgelist = edgelist.with_columns(
            txId2=pol.col("txId2").replace(self.map_id, default=-1))

        # Preparing edge_index for PyTorch
        edge_index = np.array(edgelist.to_numpy()).T
        edge_index = torch.tensor(edge_index, dtype=torch.long).contiguous()

        # Extract labels
        labels = torch.tensor(df_merge["class"].to_numpy(),
                              dtype=torch.float32).long()

        # Extract node features (with optimization)
        node_features = torch.tensor(df_merge.drop(["class",
                                                    "column_1"]).to_numpy(),
                                     dtype=torch.float32)

        num_data = node_features.shape[0]
        num_train = int(0.8 * num_data)
        num_val = int(0.1 * num_data)
        num_test = num_data - num_train - num_val

        train_index = torch.arange(num_train, dtype=torch.long)
        val_index = torch.arange(num_train,
                                 num_train + num_val,
                                 dtype=torch.long)
        test_index = torch.arange(num_train + num_val,
                                  num_train + num_val + num_test,
                                  dtype=torch.long)

        train_mask = torch.zeros_like(labels, dtype=torch.bool)
        val_mask = torch.zeros_like(labels, dtype=torch.bool)
        test_mask = torch.zeros_like(labels, dtype=torch.bool)

        train_mask[train_index] = True
        val_mask[val_index] = True
        test_mask[test_index] = True

        df = Data(
            x=node_features,
            edge_index=edge_index,
            y=labels,
            train_mask=train_mask,
            val_mask=val_mask,
            test_mask=test_mask,
        )

        self.save([df], self.processed_paths[0])



import os

from kaggle.api.kaggle_api_extended import KaggleApi


class KaggleDownloader:
    """
    A class for downloading datasets from Kaggle using the Kaggle API.

    Attributes:
        api (KaggleApi): The Kaggle API object used for authentication and downloading.
    """

    def __init__(self):
        self.api = KaggleApi()
        self.api.authenticate()

    def download_dataset(self, dataset_name, download_dir=None):
        """
        Downloads the specified dataset from Kaggle.

        Args:
            dataset_name (str): The name of the dataset on Kaggle.
            download_dir (str, optional): The directory to save the downloaded files. If not provided,
                a default directory will be created.

        Returns:
            None
        """
        if download_dir is None:
            download_dir = self.create_fintorch_data_directory()
        if not os.path.exists(download_dir):
            os.makedirs(download_dir)
        self.api.dataset_download_files(dataset_name,
                                        path=download_dir,
                                        unzip=True,
                                        quiet=False)



import multiprocessing
from typing import Callable, List, Optional, Tuple

import lightning.pytorch as pl
import numpy as np
import polars as pol
import torch
import torch_geometric.transforms as T
from huggingface_hub import hf_hub_download
from torch_geometric.data import HeteroData, InMemoryDataset
from torch_geometric.loader import LinkNeighborLoader
from tqdm import tqdm


class TransactionActorDataset(InMemoryDataset):
    """
    The Elliptic++ Data Set: Understanding Bitcoin Transactions and Actors/Wallets

    Extends the Elliptic Data Set by incorporating wallet addresses (actors) to
    enable the detection of illicit activity in the Bitcoin network.


    Purpose:

    * Maps Bitcoin transactions to real-world entities, including wallet addresses.
    * Classifies entities as licit, illicit, or unknown.
    * Facilitates the identification of illicit actors within the Bitcoin network.

    Content:

    * Inherits features and structure from the Elliptic Data Set:
        * Anonymized transaction graph from the Bitcoin blockchain.
        * 203,769 nodes (transactions) and 234,355 edges (Bitcoin flows).
        * Node features related to transactions, time, and aggregated statistics.
        * Node labels: "licit", "illicit", or "unknown".
    * Additional wallet address (actor) data:
        * 822k wallet addresses
        * Node features related to wallets, time, and other statistics.
        * Node labels: "licit", "illicit", or "unknown".

    Features:

    Here we discuss the wallet related features, for the transaction related
    features please see the elliptic.TransactionDataset.

    Note: we replaced all null values with 0 values.

    Class distribution for transactions:

    * Illicit (0)	4,545
    * Licit (1)	42,019
    * Unknown (2)	157,205

    Class distribution for wallets:

    * Illicit (0) 28,601
    * Licit (1) 338,871
    * Unknown (2) 900,788

    **Transaction related:**

    * **BTCtransacted:** Total BTC transacted (sent+received)
    * **BTCsent:** Total BTC sent
    * **BTCreceived:** Total BTC received
    * **Fees:** Total fees in BTC
    * **Feesshare:** Total fees as share of BTC transacted

    **Time related:**

    * **Blockstxs:** Number of blocks between transactions
    * **Blocksinput:** Number of blocks between being an input address
    * **Blocksoutput:** Number of blocks between being an output address
    * **Addr interactions:** Number of interactions among addresses (5 values: total, min, max, mean, median)
    * **Class:** Class label: {illicit, licit, unknown}

    **Transaction related:**

    * **Txstotal:** Total number of blockchain transactions
    * **TxSinput:** Total number of dataset transactions as input address
    * **TxSoutput:** Total number of dataset transactions as output address

    **Time related:**

    * **Timesteps:** Number of time steps transacting in
    * **Lifetime:** Lifetime in blocks
    * **Block first:** Block height first transacted in
    * **Blocklast:** Block height last transacted in
    * **Block first sent:** Block height first sent in
    * **Block first receive:** Block height first received in
    * **Repeat interactions:** Number of addresses transacted with multiple times (single value)



    Citations:

    [1] Elliptic, www.elliptic.co.
    [2] M. Weber, et al., "Anti-Money Laundering in Bitcoin: Experimenting with Graph Convolutional Networks
    for Financial Forensics", KDD ’19 Workshop on Anomaly Detection in Finance, August 2019, Anchorage, AK, USA.


    """

    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        pre_transform: Optional[Callable] = None,
        pre_filter: Optional[Callable] = None,
        force_reload: bool = False,
    ) -> None:
        super().__init__(root,
                         transform,
                         pre_transform,
                         pre_filter,
                         force_reload=force_reload)

        self.load(self.processed_paths[0])
        data = self.get(0)
        self.data, self.slices = self.collate([data])

        if force_reload:
            print("Forcing reload of the dataset")

        assert isinstance(self._data, HeteroData)

    @property
    def raw_file_names(self):
        return [
            "AddrAddr_edgelist.csv",
            "AddrTx_edgelist.csv",
            "TxAddr_edgelist.csv",
            "txs_classes.csv",
            "txs_edgelist.csv",
            "txs_features.csv",
            "wallets_features_classes_combined.csv",
        ]

    @property
    def processed_file_names(self) -> List[str]:
        """
        Returns a list of processed file names.

        Returns:
            list: A list of processed file names.
        """
        return ["transaction_actor_graph_v1.pt"]

    def download(self):
        print("Start download from HuggingFace...")
        dataset_name = "AI4FinTech/ellipticpp"
        self.downloaded_files = []
        for file in tqdm(self.raw_file_names):
            a_downloaded_files = hf_hub_download(
                repo_id=dataset_name,
                filename=file,
                repo_type="dataset",
            )
            self.downloaded_files.append(a_downloaded_files)

    def map_classes(self, df):
        """
        Maps the classes in the DataFrame to numerical values.

        Args:
            df (DataFrame): The input DataFrame containing the 'class' column.

        Returns:
            DataFrame: The modified DataFrame with the 'class' column mapped to numerical values.
        """
        # Mapping 'class' column to numerics
        # The dataset has illicit (0), licit (1), and unknown (2) entities.
        return df.with_columns(
            pol.col("class").cast(pol.Utf8).map_elements(
                lambda x: {
                    "3": 2,
                    "2": 1,
                    "1": 0
                }.get(x),
                return_dtype=pol.Int64))

    def split_data(self, num_data, splits=[0.8, 0.1]):
        """
        Splits the data into training, validation, and test sets based on the given splits.

        Args:
            num_data (int): The total number of data points.
            splits (list, optional): The split ratios for training, validation, and test sets.
                Defaults to [0.8, 0.1].

        Returns:
            tuple: A tuple containing the masks for the training, validation, and test sets.
        """

        assert len(splits) == 2, "The length of splits should be 2"
        assert sum(splits) < 1, "The sum of splits should be less than 1"

        # Generate numbers
        num_train = int(splits[0] * num_data)
        num_val = int(splits[1] * num_data)
        num_test = num_data - num_train - num_val

        # Generate ranges
        train_index = torch.arange(num_train, dtype=torch.long)
        val_index = torch.arange(num_train,
                                 num_train + num_val,
                                 dtype=torch.long)
        test_index = torch.arange(num_train + num_val,
                                  num_train + num_val + num_test,
                                  dtype=torch.long)

        # Create masks
        train_mask = torch.zeros(num_data, dtype=torch.bool)
        val_mask = torch.zeros(num_data, dtype=torch.bool)
        test_mask = torch.zeros(num_data, dtype=torch.bool)
        train_mask[train_index] = True
        val_mask[val_index] = True
        test_mask[test_index] = True

        return train_mask, val_mask, test_mask

    def prepare_edge_index(self, edgelist, mapping_dict):
        """
        Prepares the edge index for PyTorch.

        Args:
            edgelist (pandas.DataFrame): The input edge list.
            mapping_dict (dict): A dictionary containing the mapping information.

        Returns:
            torch.Tensor: The prepared edge index as a PyTorch tensor.

        """
        for k, v in mapping_dict.items():
            edgelist = edgelist.with_columns(
                pol.col(k).alias(k).map_elements(lambda x: v.get(x, x),
                                                 return_dtype=pol.Int64))

        # Preparing edge_index for PyTorch
        edgelist = np.array(edgelist.to_numpy()).T
        return torch.tensor(edgelist, dtype=torch.long).contiguous()

    def process(self):
        """
        Process the dataset by preparing features and edge_index data, and constructing a HeteroData object.

        Returns:
            None
        """

        data_list = self.raw_file_names

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        # Prepare features

        # Load transaction features and classes into a Polars DataFrame
        transaction_classes = self.map_classes(
            pol.read_csv(self.downloaded_files[3]))
        features_transaction = pol.read_csv(self.downloaded_files[5])

        features_transaction = features_transaction.join(transaction_classes,
                                                         how="left",
                                                         left_on="txId",
                                                         right_on="txId")

        # Replace null values with 0
        features_transaction = features_transaction.fill_null(0)

        # Load wallets features and classes into Polars DataFrame
        features_wallets = self.map_classes(
            pol.read_csv(self.downloaded_files[6]))
        features_wallets = features_wallets.fill_null(0)

        # Prepare labels
        transaction_labels = torch.tensor(
            features_transaction["class"].to_numpy(),
            dtype=torch.float32).long()
        wallet_labels = torch.tensor(features_wallets["class"].to_numpy(),
                                     dtype=torch.float32).long()

        # Mask data
        transaction_train_mask, transaction_val_mask, transaction_test_mask = (
            self.split_data(features_transaction.shape[0]))

        wallet_train_mask, wallet_val_mask, wallet_test_mask = self.split_data(
            features_wallets.shape[0])

        # PyTorch features
        tensor_features_wallets = torch.tensor(
            features_wallets.drop(["class", "Time step",
                                   "address"]).to_numpy(),
            dtype=torch.float32,
        )

        tensor_features_transaction = torch.tensor(
            features_transaction.drop(["class", "Time step",
                                       "txId"]).to_numpy(),
            dtype=torch.float32,
        )

        # Prepare edge_index data

        # Replace nodes with indices for transactions and wallets
        features_transaction = features_transaction.with_columns(
            mapped_id=pol.arange(0, features_transaction.shape[0]))

        features_wallets = features_wallets.with_columns(
            mapped_id=pol.arange(0, features_wallets.shape[0]))

        wallets_mapping = dict(
            zip(features_wallets["address"], features_wallets["mapped_id"]))
        transaction_mapping = dict(
            zip(features_transaction["txId"],
                features_transaction["mapped_id"]))

        edgelist_addr_addr = pol.read_csv(self.downloaded_files[0])
        edgelist_addr_tx = pol.read_csv(self.downloaded_files[1])
        edgelist_tx_addr = pol.read_csv(self.downloaded_files[2])
        edgelist_tx_tx = pol.read_csv(self.downloaded_files[4])

        tx_tx_dict = {
            "txId1": transaction_mapping,
            "txId2": transaction_mapping
        }
        addr_addr_dict = {
            "input_address": wallets_mapping,
            "output_address": wallets_mapping,
        }
        addr_tx_dict = {
            "input_address": wallets_mapping,
            "txId": transaction_mapping
        }
        tx_addr_dict = {
            "txId": transaction_mapping,
            "output_address": wallets_mapping
        }

        addr_tx_edge_index = self.prepare_edge_index(edgelist_addr_tx,
                                                     addr_tx_dict)
        tx_addr_edge_index = self.prepare_edge_index(edgelist_tx_addr,
                                                     tx_addr_dict)
        addr_addr_edge_index = self.prepare_edge_index(edgelist_addr_addr,
                                                       addr_addr_dict)
        tx_tx_edge_index = self.prepare_edge_index(edgelist_tx_tx, tx_tx_dict)

        # Construct HeteroData object and store it

        df = HeteroData(
            wallets={
                "x": tensor_features_wallets,
                "y": wallet_labels,
                "train_mask": wallet_train_mask,
                "val_mask": wallet_val_mask,
                "test_mask": wallet_test_mask,
            },
            transactions={
                "x": tensor_features_transaction,
                "y": transaction_labels,
                "train_mask": transaction_train_mask,
                "val_mask": transaction_val_mask,
                "test_mask": transaction_test_mask,
            },
        )

        df["transactions", "transactions"].edge_index = tx_tx_edge_index
        df["transactions", "wallets"].edge_index = tx_addr_edge_index
        df["wallets", "transactions"].edge_index = addr_tx_edge_index
        df["wallets", "wallets"].edge_index = addr_addr_edge_index

        self.save([df], self.processed_paths[0])


class EllipticppDataModule(pl.LightningDataModule):
    """
    LightningDataModule for handling the Elliptic++ dataset. The EllipticDataModule class is responsible for loading,
    preprocessing, and preparing the Elliptic++ dataset for use in Lightning Model Modules.
    It handles tasks like splitting the data into training, validation, and testing sets, generating data loaders,
    and implementing neighbor sampling strategies for efficient training.


    Args:
        edge (Tuple[str, str, str]): A tuple representing the edge types in the dataset.
        num_val (float, optional): The ratio of validation data to the total dataset. Defaults to 0.1.
        num_test (float, optional): The ratio of test data to the total dataset. Defaults to 0.1.
        disjoint_train_ratio (float, optional): The ratio of disjoint training data to the total dataset.
                                                Defaults to 0.3.
        neg_sampling_ratio (float, optional): The ratio of negative samples to positive samples. Defaults to 2.0.
        num_neighbors (List[int], optional): A list of integers representing the number of neighbors to consider.
                                             Defaults to [10, 30].
        batch_size (int, optional): The batch size for data loading. Defaults to 128.
        neg_sampling (str, optional): The type of negative sampling. Defaults to "binary".
        num_workers (int, optional): The number of workers for data loading. Defaults to -1.

    """

    def __init__(
        self,
        edge: Tuple[str, str, str],
        num_val: float = 0.1,
        num_test: float = 0.1,
        disjoint_train_ratio: float = 0.3,
        neg_sampling_ratio: float = 2.0,
        num_neighbors: List[int] = [10, 30],
        batch_size: int = 128,
        neg_sampling: str = "binary",
        num_workers: int = -1,
        force_reload=False,
    ) -> None:
        super().__init__()

        assert (isinstance(edge, tuple)
                and len(edge) == 3), "edge must be a tuple of length 3"
        assert (isinstance(num_val, float) and
                0 <= num_val <= 1), "num_val must be a float between 0 and 1"
        assert (isinstance(num_test, float) and
                0 <= num_test <= 1), "num_test must be a float between 0 and 1"
        assert (isinstance(disjoint_train_ratio, float)
                and 0 <= disjoint_train_ratio <= 1
                ), "disjoint_train_ratio must be a float between 0 and 1"
        assert (isinstance(neg_sampling_ratio, float) and neg_sampling_ratio
                > 0), "neg_sampling_ratio must be a positive float"
        assert isinstance(num_neighbors, list) and all(
            isinstance(n, int)
            for n in num_neighbors), "num_neighbors must be a list of integers"
        assert (isinstance(batch_size, int)
                and batch_size > 0), "batch_size must be a positive integer"
        assert isinstance(neg_sampling, str), "neg_sampling must be a string"

        self.edge = edge
        self.num_val = num_val
        self.num_test = num_test
        self.disjoint_train_ratio = disjoint_train_ratio
        self.neg_sampling_ratio = neg_sampling_ratio
        self.num_neighbors = num_neighbors
        self.batch_size = batch_size
        self.neg_sampling = neg_sampling
        self.force_reload = force_reload

        if num_workers == -1:
            self.num_workers = multiprocessing.cpu_count() - 1
        else:
            self.num_workers = num_workers

    def setup(self, stage=None):
        dataset = TransactionActorDataset("~/.fintorch_data",
                                          force_reload=self.force_reload)
        self.dataset = dataset[0]
        self.split_dataset(dataset[0])

    def split_dataset(self, dataset):
        """
        Splits the given dataset into training, validation, and test sets and makes them available
        as self.train_data, self.val_data, and self.test_data

        Args:
            dataset: The input dataset to be split.

        Returns:
            None
        """

        transform = T.RandomLinkSplit(
            num_val=0.1,
            num_test=0.1,
            disjoint_train_ratio=0.3,
            neg_sampling_ratio=2.0,
            edge_types=[
                ("wallets", "to", "transactions"),
                ("transactions", "to", "wallets"),
                ("wallets", "to", "wallets"),
                ("transactions", "to", "transactions"),
            ],
        )
        self.train_data, self.val_data, self.test_data = transform(dataset)

    def train_dataloader(self):
        """
        Returns a DataLoader object for training.
        """
        src, to, dst = self.edge

        loader = LinkNeighborLoader(
            self.train_data,
            num_neighbors=self.num_neighbors,
            batch_size=self.batch_size,
            shuffle=True,
            edge_label_index=(
                (src, to, dst),
                self.train_data[src, to, dst].edge_index,
            ),
            neg_sampling=self.neg_sampling,
            num_workers=self.num_workers,
        )

        return loader

    def val_dataloader(self):
        """
        Returns a data loader for the validation data.

        Returns:
            loader (LinkNeighborLoader): A data loader for the validation data.
        """
        src, to, dst = self.edge

        loader = LinkNeighborLoader(
            self.val_data,
            num_neighbors=self.num_neighbors,
            batch_size=self.batch_size,
            edge_label_index=(
                (src, to, dst),
                self.val_data[src, to, dst].edge_index,
            ),
            neg_sampling=self.neg_sampling,
            num_workers=self.num_workers,
        )

        return loader

    def test_dataloader(self):
        """
        Returns a DataLoader object for testing the dataset.

        Returns:
            DataLoader: A DataLoader object that loads test data for the dataset.
        """
        src, to, dst = self.edge
        loader = LinkNeighborLoader(
            self.test_data,
            num_neighbors=self.num_neighbors,
            batch_size=self.batch_size,
            edge_label_index=(
                (src, to, dst),
                self.test_data[src, to, dst].edge_index,
            ),
            neg_sampling=self.neg_sampling,
            num_workers=self.num_workers,
        )

        return loader



from typing import Dict

import lightning as L
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch_geometric.nn as nng
import torchmetrics
from torch_geometric.nn import SAGEConv


def GraphBEANLoss(feature_predictions, edge_predictions,
                  ground_truth_sampled_data, edge):
    """
    Calculates the loss function for the GraphBEAN model.

    Args:
        feature_predictions (dict): A dictionary containing the predicted features for each node type.
        edge_predictions (torch.Tensor): The predicted edge values.
        ground_truth_sampled_data (dict): A dictionary containing the ground truth data for each node type.
        edge (tuple): A tuple representing the source, target, and destination nodes for the edge.

    Returns:
        torch.Tensor: The total loss value.

    """
    # Loss of the feature reconstruction per node type
    feature_loss = 0
    for key in feature_predictions.keys():
        feature_loss += nn.MSELoss()(feature_predictions[key],
                                     ground_truth_sampled_data[key].x)

    # Edge prediction loss (one edge type)
    src, to, dst = edge
    edge_loss = F.binary_cross_entropy_with_logits(
        edge_predictions, ground_truth_sampled_data[src, to, dst].edge_label)
    # Total loss function
    total_loss = feature_loss + edge_loss
    return total_loss


def GraphBEANLossClassifier(
    feature_predictions,
    edge_predictions,
    ground_truth_sampled_data,
    edge,
    node_pred,
    node_ground_truth,
):
    """
    Calculates the total loss for the GraphBEAN model with a classifier.
    We assume 3 classes:
    * 0/1: two classes of interest
    * 2: the class we exlcude (for example unlabelled data)

    Args:
        feature_predictions (Tensor): Predictions of the node features.
        edge_predictions (Tensor): Predictions of the edge features.
        ground_truth_sampled_data (Tensor): Ground truth data.
        edge (Tensor): Edge tensor.
        node_pred (Tensor): Predictions of the node labels.
        node_ground_truth (Tensor): Ground truth node labels.

    Returns:
        Tensor: The total loss for the GraphBEAN model with a classifier.
    """

    # Loss of the feature reconstruction per node type
    loss = GraphBEANLoss(feature_predictions, edge_predictions,
                         ground_truth_sampled_data, edge)

    classification_loss_fn = torch.nn.CrossEntropyLoss()

    # We assume that we have classes 1/2/3 where 3 is unlabbeled. We filter on 3.
    # Define more elegant solution see issue

    idx = node_ground_truth != 2
    filtered = node_ground_truth[idx]

    class_loss = classification_loss_fn(node_pred[idx, :], filtered)

    weight = 3  # weight > 1, then class loss is weighted more
    # Harmonic mean of feature_loss and total_loss
    harmonic_mean = (1 + weight) * (loss * class_loss) / (weight * loss +
                                                          class_loss)
    total_loss = harmonic_mean

    return total_loss


class StructureDecoder(nn.Module):
    """
    Structure decoder module for the graphBEAN model.

    Args:
        in_channels (int): Number of input channels.
        hidden_channels (int): Number of hidden channels.
        out_channels (int): Number of output channels.
        num_layers (int): Number of layers in the MLP.

    Attributes:
        mlp_layers_src (nn.ModuleList): List of MLP layers for source node features.
        mlp_layers_dst (nn.ModuleList): List of MLP layers for destination node features.
    """

    def __init__(self, in_channels, hidden_channels, out_channels, num_layers):
        super().__init__()

        assert num_layers >= 2

        self.mlp_layers_src = nn.ModuleList()
        self.mlp_layers_src.append(nn.Linear(in_channels, hidden_channels))
        for _ in range(num_layers - 2):
            self.mlp_layers_src.append(
                nn.Linear(hidden_channels, hidden_channels))
        self.mlp_layers_src.append(nn.Linear(hidden_channels, out_channels))

        self.mlp_layers_dst = nn.ModuleList()
        self.mlp_layers_dst.append(nn.Linear(in_channels, hidden_channels))
        for _ in range(num_layers - 2):
            self.mlp_layers_dst.append(
                nn.Linear(hidden_channels, hidden_channels))
        self.mlp_layers_dst.append(nn.Linear(hidden_channels, out_channels))

    def forward(self, src, dst, edge_label_index):
        """
        Performs the forward pass of the StructureDecoder model.

        Args:
            src (torch.Tensor): Source node features.
            dst (torch.Tensor): Destination node features.
            edge_label_index (tuple): Tuple containing the indices of the edge labels.

        Returns:
            torch.Tensor: The result of the forward pass.
        """
        x_src_features = src[edge_label_index[0]]
        x_dst_features = dst[edge_label_index[1]]

        # Loop over the self.mlp_layers_src
        src = x_src_features
        for layer in self.mlp_layers_src:
            src = layer(src)

        # Loop over the self.mlp_layers_dst
        dst = x_dst_features
        for layer in self.mlp_layers_dst:
            dst = layer(dst)

        return torch.sigmoid((src * dst).sum(dim=-1))


class GraphBEAN(nn.Module):
    """
    GraphBEAN model for interaction-focused anomaly detection on bipartite node-and-edge-attributed graphs.

    Reference:
    Fathony, R., Ng, J., & Chen, J. (2023, June). Interaction-focused anomaly detection on bipartite
    node-and-edge-attributed graphs. In 2023 International Joint Conference on Neural Networks (IJCNN) (pp. 1-10). IEEE.

    Args:
        n_encoder_layers (int): Number of encoder layers.
        n_feature_decoder_layers (int): Number of feature decoder layers.
        hidden_channels (int): Number of hidden channels.
        features_channels (dict): Dictionary mapping edge types to feature channels.
        structure_decoder_head_out_channels (int): Number of output channels for the structure decoder.
        structure_decoder_head_layers (int): Number of layers for the structure decoder.
        conv_type (callable): Convolutional layer type.
        edge_types: List of edge types.
        aggr (str, optional): Aggregation method for the convolutional layers. Defaults to 'sum'.

    Attributes:
        encoder_layers (nn.ModuleList): List of encoder layers.
        decoder_layers (nn.ModuleList): List of feature decoder layers.
        graph_edge_prediction (StructureDecoder): Graph classifier head.

    """

    def __init__(
        self,
        n_encoder_layers: int,
        n_feature_decoder_layers: int,
        hidden_channels: int,
        features_channels: dict,
        structure_decoder_head_out_channels: int,
        structure_decoder_head_layers: int,
        conv_type: callable,
        edge_types,
        aggr: str = "sum",
    ):
        super().__init__()

        assert n_encoder_layers > 0, "Number of encoder layers must be greater than 0."
        assert (
            n_feature_decoder_layers
            > 0), "Number of feature decoder layers must be greater than 0."
        assert hidden_channels > 0, "Number of hidden channels must be greater than 0."

        # Encoder layers
        self.encoder_layers = nn.ModuleList()
        first_encoder_layer = nng.HeteroConv(
            {
                edge_type: conv_type(-1, hidden_channels)
                for edge_type in edge_types
            },
            aggr=aggr,
        )

        self.encoder_layers.append(first_encoder_layer)

        for _ in range(n_encoder_layers - 1):
            self.encoder_layers.append(
                nng.HeteroConv(
                    {
                        edge_type: conv_type(hidden_channels, hidden_channels)
                        for edge_type in edge_types
                    },
                    aggr=aggr,
                ))  # Pass conv_type argument

        # FeatureDecoder layers
        self.decoder_layers = nn.ModuleList()
        for _ in range(n_feature_decoder_layers):
            self.decoder_layers.append(
                nng.HeteroConv(
                    {
                        edge_type: conv_type(hidden_channels, hidden_channels)
                        for edge_type in edge_types
                    },
                    aggr=aggr,
                ))

        decoder_last_conv_layer = nng.HeteroConv(
            {
                edge_type:
                conv_type(hidden_channels, features_channels[edge_type[2]])
                for edge_type in edge_types
            },
            aggr=aggr,
        )

        self.decoder_layers.append(decoder_last_conv_layer)

        # Graph classifier head
        self.graph_edge_prediction = StructureDecoder(
            hidden_channels,
            hidden_channels,
            structure_decoder_head_out_channels,
            structure_decoder_head_layers,
        )

    def forward(self, data, edge):
        """
        Forward pass of the GraphBEAN model.

        Args:
            data (Tensor): The input data.
            edge (Tuple): Tuple containing the source, target, and destination nodes.

        Returns:
            Tuple: A tuple containing the following elements:
                - None: Placeholder for future use (classification model).
                - hidden_representation (Tensor): The hidden representation obtained from the encoder layers.
                - feature_out (Tensor): The output of the feature decoding layers.
                - edge_prediction (Tensor): The predicted edge labels.
        """

        # loop through the encoder layers to obtain the hidden representation
        hidden_representation = data.x_dict
        for layer in self.encoder_layers:
            hidden_representation = layer(
                hidden_representation,
                data.edge_index_dict,
            )
            hidden_representation = {
                key: F.relu(x)
                for key, x in hidden_representation.items()
            }

        # Obtain the feature decoding output after looping through the feature decoding layers
        feature_out = hidden_representation
        for layer in self.decoder_layers:
            feature_out = layer(
                feature_out,
                data.edge_index_dict,
            )
            feature_out = {key: F.relu(x) for key, x in feature_out.items()}

        src, to, dst = edge

        edge_prediction = self.graph_edge_prediction(
            hidden_representation[src],
            hidden_representation[dst],
            data[src, to, dst].edge_label_index,
        )

        return None, hidden_representation, feature_out, edge_prediction


class GraphBeanClassifier(nn.Module):
    """
    GraphBEAN model for interaction-focused anomaly detection on bipartite node-and-edge-attributed graphs.
    We ammended the original implementation to include a classification head such that the model works for
    semi-supervised settings.

    Reference:
    Fathony, R., Ng, J., & Chen, J. (2023, June). Interaction-focused anomaly detection on bipartite
    node-and-edge-attributed graphs. In 2023 International Joint Conference on Neural Networks (IJCNN) (pp. 1-10). IEEE.


    Args:
        hetero_data (torch_geometric.data.HeteroData): The heterogeneous input graph data.
        n_encoder_layers (int): The number of layers in the encoder.
        n_feature_decoder_layers (int): The number of layers in the feature decoder.
        hidden_channels (int): The number of hidden channels in the model.
        features_channels (Dict): A dictionary specifying the number of channels for each feature type.
        structure_decoder_head_out_channels (int): The number of output channels in the structure decoder head.
        structure_decoder_head_layers (int): The number of layers in the structure decoder head.
        class_head_layers (int, optional): The number of layers in the classifier head. Defaults to 1.
        classes (int, optional): The number of classes for classification. Defaults to 3.
        conv_type (callable, optional): The type of convolutional layer to use. Defaults to SAGEConv.
        edge_types (list, optional): A list specifying the types of edges in the graph. Defaults to None.
        aggr (str, optional): The aggregation method to use. Defaults to 'sum'.

    Attributes:
        data_types (list): A list of the data types in the heterogeneous graph.
        graph_bean_layer (GraphBEAN): The GraphBEAN layer.
        classifierHead (nn.ModuleList): The classifier head module list.

    """

    def __init__(
        self,
        hetero_data,
        n_encoder_layers: int,
        n_feature_decoder_layers: int,
        hidden_channels: int,
        features_channels: Dict,
        structure_decoder_head_out_channels: int,
        structure_decoder_head_layers: int,
        class_head_layers: int = 1,
        classes: int = 3,
        conv_type: callable = SAGEConv,
        edge_types=None,
        aggr="sum",
    ):

        super().__init__()

        self.data_types = hetero_data.metadata()[0]

        self.graph_bean_layer = GraphBEAN(
            n_encoder_layers,
            n_feature_decoder_layers,
            hidden_channels,
            features_channels,
            structure_decoder_head_out_channels,
            structure_decoder_head_layers,
            conv_type,
            edge_types,
            aggr,
        )

        self.classifierHead = nn.ModuleList([
            nng.HeteroDictLinear(hidden_channels, hidden_channels,
                                 hetero_data.metadata()[0])
        ])

        for _ in range(class_head_layers):
            self.classifierHead.append(
                nng.HeteroDictLinear(hidden_channels, hidden_channels,
                                     hetero_data.metadata()[0]))
        self.classifierHead.append(
            nng.HeteroDictLinear(hidden_channels, classes,
                                 hetero_data.metadata()[0]))

    def forward(self, data, edge):
        """
        Forward pass of the GraphBeanClassifier.

        Args:
            data (torch_geometric.data.Data): The input graph data.
            edge (torch.Tensor): The edge indices of the graph.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]: The output tensors from the forward pass.
                - class_probs (torch.Tensor): The class probabilities.
                - hidden_representation (torch.Tensor): The hidden representation.
                - feature_out (torch.Tensor): The output features.
                - edge_prediction (torch.Tensor): The edge predictions.

        """

        _, hidden_representation, feature_out, edge_prediction = self.graph_bean_layer(
            data, edge)

        class_probs = hidden_representation
        for layer in self.classifierHead:
            class_probs = layer(class_probs)

        output = {}
        for key in self.data_types:
            output[key] = torch.sigmoid(class_probs[key])

        return class_probs, hidden_representation, feature_out, edge_prediction


class GraphBEANModule(L.LightningModule):
    """
    LightningModule implementation for the GraphBEAN model.

    Args:
        edge: The input edge.
        edge_types: The types of edges in the graph.
        loss_fn: The loss function to use. Defaults to None.
        learning_rate: The learning rate for the optimizer. Defaults to 0.01.
        encoder_layers: The number of layers in the encoder. Defaults to 2.
        decoder_layers: The number of layers in the decoder. Defaults to 2.
        hidden_layers: The number of hidden layers. Defaults to 128.
        structure_decoder_head_out_channels: The number of output channels in the structure decoder head.
        Defaults to 50.
        structure_decoder_head_layers: The number of layers in the structure decoder head. Defaults to 10.
        classifier: Whether the model is a classifier. Defaults to False.
        class_head_layers: The number of layers in the class head. Defaults to 3.
        classes: The number of classes. Defaults to 2.
        predict: The attribute to predict. Defaults to None.
        data: The dataset. Defaults to None.
        conv_type: The type of convolutional layer to use. Defaults to SAGEConv.
        aggr: The aggregation method for the convolutional layer. Defaults to 'sum'.
    """

    def __init__(
        self,
        edge,
        edge_types,
        loss_fn=None,
        learning_rate=0.01,
        encoder_layers=2,
        decoder_layers=2,
        hidden_layers=128,
        structure_decoder_head_out_channels=50,
        structure_decoder_head_layers=10,
        classifier=False,
        class_head_layers=3,
        classes=2,
        predict=None,
        data=None,
        conv_type=SAGEConv,
        aggr="sum",
    ):
        super().__init__()
        self.edge = edge
        self.predict = predict

        if classifier:
            assert predict is not None

        if data:
            self.dataset = data
        else:
            self.dataset = None

        if classifier:
            self.loss_fn = loss_fn if loss_fn is not None else GraphBEANLossClassifier
        else:
            self.loss_fn = loss_fn if loss_fn is not None else GraphBEANLoss

        self.learning_rate = learning_rate
        self.encoder_layers = encoder_layers
        self.decoder_layers = decoder_layers
        self.hidden_layers = hidden_layers
        self.classifier = classifier
        self.class_head_layers = class_head_layers
        self.classes = classes
        self.conv_type = conv_type
        self.structure_decoder_head_out_channels = structure_decoder_head_out_channels
        self.structure_decoder_head_layers = structure_decoder_head_layers
        self.aggr = aggr

        self.edge_types = edge_types
        if isinstance(self.edge_types, str):
            self.edge_types = [self.edge_types]

        # Convert edge_type strings to tuples
        self.edge_types = [
            tuple(edge_type.split("_"))
            if isinstance(edge_type, str) else edge_type
            for edge_type in self.edge_types
        ]

        self.accuracy = torchmetrics.classification.Accuracy(
            task="multiclass", num_classes=classes, average="macro")

        self.f1 = torchmetrics.classification.F1Score(task="multiclass",
                                                      num_classes=classes,
                                                      average="macro")

        self.recall = torchmetrics.classification.Recall(task="multiclass",
                                                         num_classes=classes,
                                                         average="macro")

        self.precision = torchmetrics.classification.Precision(
            task="multiclass", num_classes=classes, average="macro")

        self.confmat = torchmetrics.classification.ConfusionMatrix(
            task="multiclass", num_classes=classes)

        self.aucroc = torchmetrics.classification.AUROC(task="multiclass",
                                                        num_classes=classes,
                                                        average="macro")

        self.validation_step_outputs = []

        self.save_hyperparameters()

    def setup(self, stage=None):
        """
        Function called before validate and fit.
        Sets up the model and optimizers based on the provided dataset and configuration.

        Args:
            stage: Optional[str], the current stage of training. Defaults to None.

        Returns:
            None
        """

        # Get the dataset from the datamodule
        if self.dataset is None:
            self.dataset = self.trainer.datamodule.dataset

        # Required for the feature encoder-decoder
        mapping = dict()
        for key in self.dataset.metadata()[0]:
            mapping[key] = self.dataset[key].x.shape[1]

        if self.classifier:
            # Initialize your model using the dataset
            self.model = GraphBeanClassifier(
                self.dataset,
                self.encoder_layers,
                self.decoder_layers,
                self.hidden_layers,
                mapping,
                self.structure_decoder_head_out_channels,
                self.structure_decoder_head_layers,
                self.class_head_layers,
                self.classes,
                self.conv_type,
                self.edge_types,
                self.aggr,
            )

        else:
            # Initialize your model using the dataset
            self.model = GraphBEAN(
                self.encoder_layers,
                self.decoder_layers,
                self.hidden_layers,
                mapping,
                self.structure_decoder_head_out_channels,
                self.structure_decoder_head_layers,
                self.conv_type,
                self.edge_types,
                self.aggr,
            )

        self.optimizers = optim.Adam(self.model.parameters(),
                                     lr=self.learning_rate)

    def forward(self, batch, edge):
        """
        Forward pass of the graphBEAN model.

        Args:
            batch: The input batch.
            edge: The input edge.

        Returns:
            The output of the model.
        """
        return self.model(batch, edge)

    def loss(self, batch, class_probs, pred_features, pred_edges):
        """
        Computes the loss for the given batch.

        Args:
            batch: The input batch.
            class_probs: The class probabilities.
            pred_features: The predicted features.
            pred_edges: The predicted edges.

        Returns:
            The computed loss.
        """
        if self.classifier:
            loss = self.loss_fn(
                pred_features,
                pred_edges,
                batch,
                self.edge,
                class_probs[self.predict],
                batch[self.predict].y,
            )
        else:
            loss = self.loss_fn(pred_features, pred_edges, batch, self.edge)

        return loss

    def training_step(self, batch, batch_idx):
        """
        Performs a single training step on the given batch of data.

        Args:
            batch: The input batch of data.
            batch_idx: The index of the current batch.

        Returns:
            The loss value computed during the training step.
        """
        class_probs, _, pred_features, pred_edges = self(batch, self.edge)

        loss = self.loss(batch, class_probs, pred_features, pred_edges)

        self.log("train_loss", loss, on_step=True, prog_bar=True)

        if self.classifier:
            node_ground_truth = batch[self.predict].y
            # print(f"unique:{batch[self.predict].y.unique()}")
            idx = node_ground_truth != 2
            batch_subset = node_ground_truth[idx]
            # print(f"batch:{batch[self.predict].y.shape} sub:{sum(idx)}")
            output_class = torch.argmax(class_probs[self.predict],
                                        dim=1).long()
            output_class_subset = output_class[idx]
            self.accuracy(output_class_subset, batch_subset)
            self.log(
                "train_accuracy",
                self.accuracy,
                prog_bar=True,
                on_step=True,
                on_epoch=False,
            )

        return loss

    def validation_step(self, batch, batch_idx):
        """
        Performs a validation step on the given batch of data.

        Args:
            batch: The batch of data for validation.
            batch_idx: The index of the current batch.

        Returns:
            None
        """

        class_probs, _, pred_features, pred_edges = self(batch, self.edge)
        loss = self.loss(batch, class_probs, pred_features, pred_edges)

        self.log(
            "val_loss",
            loss,
            on_step=False,
            on_epoch=True,
            prog_bar=False,
            batch_size=64,
        )

        if self.classifier:
            node_ground_truth = batch[self.predict].y
            idx_filter = node_ground_truth != 2
            batch_subset = node_ground_truth[idx_filter]
            output_class = torch.argmax(class_probs[self.predict],
                                        dim=1).long()
            output_class_subset = output_class[idx_filter]
            self.accuracy(output_class_subset, batch_subset)
            self.log(
                "val_acc_step",
                self.accuracy,
                on_step=False,
                on_epoch=True,
                prog_bar=True,
                batch_size=64,
            )
            self.f1(output_class_subset, batch_subset)
            self.log(
                "val_f1",
                self.f1,
                on_step=False,
                on_epoch=True,
                prog_bar=True,
                batch_size=64,
            )
            self.recall(output_class_subset, batch_subset)
            self.log(
                "val_recall",
                self.recall,
                on_step=False,
                on_epoch=True,
                prog_bar=True,
                batch_size=64,
            )
            self.precision(output_class_subset, batch_subset)
            self.log(
                "val_precision",
                self.precision,
                on_step=False,
                on_epoch=True,
                prog_bar=True,
                batch_size=64,
            )

    def on_validation_epoch_end(self):
        """
        Performs operations at the end of each validation epoch.

        This method clears the validation step outputs.

        Returns:
            None
        """

        self.validation_step_outputs.clear()

    def test_step(self, batch, batch_idx):
        """
        Performs a test step on the given batch of data.

        Args:
            batch: The batch of data for validation.
            batch_idx: The index of the current batch.

        Returns:
            None
        """

        class_probs, _, pred_features, pred_edges = self(batch, self.edge)
        self.loss(batch, class_probs, pred_features, pred_edges)

        if self.classifier:
            node_ground_truth = batch[self.predict].y
            idx_filter = node_ground_truth != 2
            batch_subset = node_ground_truth[idx_filter]
            output_class = torch.argmax(class_probs[self.predict],
                                        dim=1).long()
            output_class_subset = output_class[idx_filter]
            self.accuracy(output_class_subset, batch_subset)
            self.log(
                "test_acc",
                self.accuracy,
                on_step=False,
                on_epoch=True,
                prog_bar=True,
                batch_size=64,
            )
            self.f1(output_class_subset, batch_subset)
            self.log(
                "test_f1",
                self.f1,
                on_step=False,
                on_epoch=True,
                prog_bar=True,
                batch_size=64,
            )
            self.recall(output_class_subset, batch_subset)
            self.log(
                "test_recall",
                self.recall,
                on_step=False,
                on_epoch=True,
                prog_bar=True,
                batch_size=64,
            )
            self.precision(output_class_subset, batch_subset)
            self.log(
                "test_precision",
                self.precision,
                on_step=False,
                on_epoch=True,
                prog_bar=True,
                batch_size=64,
            )

    def configure_optimizers(self):
        """
        Configures the optimizers for the model.

        Returns:
            optimizers: The configured optimizers for the model.
        """
        return self.optimizers



#!/usr/bin/env python
"""The setup script."""

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "Click>=7.0",
    "torch",
    "torch_geometric",
    "lightning",
    "kaggle",
    "polars",
    "numpy",
    "huggingface_hub",
    "seaborn",
    "networkx",
    "optuna",
    "wandb>=0.12.1",
    "optuna-integration",
]

test_requirements = [
    "pytest>=3",
]

setup(
    author="Marcel Boersma",
    author_email="boersma.marcel@gmail.com",
    python_requires=">=3.10",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.10",
    ],
    description="AI4FinTech project repository",
    entry_points={
        "console_scripts": [
            "fintorch=fintorch.cli:fintorch",
            "fintrainer=fintorch.cli:trainer",
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type='text/x-rst',
    include_package_data=True,
    keywords="fintorch",
    name="fintorch",
    packages=find_packages(include=["fintorch", "fintorch.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/AI4FinTech/fintorch",
    version="0.1.8",
    zip_safe=False,
)
